// PD_RENDEZVOUS_HOST
// PD_RENDEZVOUS_PORT

// PD_INSPECTION_ADDRESS
// PD_INSPECTION_PORT

// PD_PRIVATE_KEY
// PD_AUTHORIZED_KEY

// PD_LOCAL_SSH_PORT

package main

import (
	"flag"
	"fmt"
	"io"
	"log"
	"net"
	"net/url"
	"os"
	"os/exec"
	"strconv"
	"strings"
	"syscall"
	"unsafe"

	"github.com/creack/pty"
	gssh "github.com/gliderlabs/ssh"
	"golang.org/x/crypto/ssh"
)

type Endpoint struct {
	User string
	Host string
	Port int
}

func (endpoint *Endpoint) String() string {
	return fmt.Sprintf("%s@%s:%d", endpoint.User, endpoint.Host, endpoint.Port)
}

func (endpoint *Endpoint) StringIP() string {
	return fmt.Sprintf("%s:%d", endpoint.Host, endpoint.Port)
}

// From https://sosedoff.com/2015/05/25/ssh-port-forwarding-with-go.html
// Handle local client connections and tunnel data to the remote server
// Will use io.Copy - http://golang.org/pkg/io/#Copy
func handleClient(client net.Conn, remote net.Conn) {
	defer client.Close()
	chDone := make(chan bool)

	// Start remote -> local data transfer
	go func() {
		_, err := io.Copy(client, remote)
		if err != nil {
			log.Println(fmt.Sprintf("error while copy remote->local: %s", err))
		}
		chDone <- true
	}()

	// Start local -> remote data transfer
	go func() {
		_, err := io.Copy(remote, client)
		if err != nil {
			log.Println(fmt.Sprintf("error while copy local->remote: %s", err))
		}
		chDone <- true
	}()

	<-chDone
}

var privateKey ssh.AuthMethod
var gprivateKey []byte
var authorizedKey gssh.PublicKey
var rendezvousUrl *url.URL

// local service to be forwarded
var localEndpoint = Endpoint{
	User: "debugci",
	Host: "127.0.0.1",
	Port: 2222,
}

// remote SSH server
var rendezvousEndpoint Endpoint

// remote forwarding port (on remote SSH server network)
var inspectionEndpoint = Endpoint{
	User: "debugci",
	Host: "127.0.0.1",
	Port: 8123,
}

var rendezvous = flag.String("rendezvous", "", "user@host:port of the rendezvous place (DCI_RENDEZVOUS)")
var privateKeyFile = flag.String("private-key-file", "", "Private key to get to rendezvous place and server (DCI_PRIVATE_KEY)")
var authorizedKeyFile = flag.String("authorized-key-file", "", "Public key of authorized client (DCI_AUTHORIZED_KEY)")

func parseArguments() {
	flag.Parse()

	var err error
	if *rendezvous == "" {
		*rendezvous = os.Getenv("DCI_RENDEZVOUS")
	}

	*rendezvous = "ssh://" + *rendezvous
	rendezvousUrl, err = url.Parse(*rendezvous)

	portStr := rendezvousUrl.Port()
	port := 22
	if portStr != "" {
		port, err = strconv.Atoi(portStr)
		if err != nil {
			log.Fatalf("Port must be numeric or empty: %s\n", portStr)
		}
	}

	rendezvousEndpoint = Endpoint{
		User: rendezvousUrl.User.String(),
		Host: rendezvousUrl.Hostname(),
		Port: port,
	}

	// Private key
	var keyBuffer []byte
	if *privateKeyFile != "" {
		log.Printf("Reading private key from file %s\n", *privateKeyFile)
		keyBuffer, err = os.ReadFile(*privateKeyFile)
		if err != nil {
			log.Fatalf("Cannot read SSH private key file %s\n", *privateKeyFile)
		}
	} else {
		log.Println("Reading private key from env")
		keyBuffer = []byte(os.Getenv("DCI_PRIVATE_KEY"))
	}
	gprivateKey = keyBuffer
	key, err := ssh.ParsePrivateKey(keyBuffer)

	if err != nil {
		log.Fatalln("Cannot parse SSH private key file")
	}
	privateKey = ssh.PublicKeys(key)

	// Authorized key
	if *authorizedKeyFile != "" {
		log.Printf("Reading authorized key from file %s\n", *authorizedKeyFile)
		keyBuffer, err = os.ReadFile(*authorizedKeyFile)
		if err != nil {
			log.Fatalf("Cannot read SSH public key file %s\n", *authorizedKeyFile)
		}
	} else {
		log.Println("Reading authorized key from env")
		keyBuffer = []byte(os.Getenv("DCI_AUTHORIZED_KEY"))
	}
	authorizedKey, _, _, _, err = gssh.ParseAuthorizedKey(keyBuffer)
	if err != nil {
		log.Fatalf("Cannot parse authorized key: %s\n", keyBuffer)
	}
}

func main() {
	parseArguments()

	go server()

	// refer to https://godoc.org/golang.org/x/crypto/ssh for other authentication types
	sshConfig := &ssh.ClientConfig{
		// SSH connection username
		User: rendezvousEndpoint.User,
		Auth: []ssh.AuthMethod{
			privateKey,
		},
		HostKeyCallback: ssh.InsecureIgnoreHostKey(),
	}

	// Connect to SSH remote server using rendezvousEndpoint
	serverConn, err := ssh.Dial("tcp", rendezvousEndpoint.StringIP(), sshConfig)
	if err != nil {
		log.Fatalf("Dial INTO remote server error: %s\n", err)
	}

	// Listen on remote server port
	listener, err := serverConn.Listen("tcp", inspectionEndpoint.StringIP())
	if err != nil {
		log.Fatalf("Listen open port ON remote server error: %s\n", err)
	}
	defer listener.Close()

	// handle incoming connections on reverse forwarded tunnel
	for {
		// Open a (local) connection to localEndpoint whose content will be forwarded so rendezvousEndpoint
		local, err := net.Dial("tcp", localEndpoint.StringIP())
		log.Println("New connection")
		if err != nil {
			log.Fatalf("Dial INTO local service error: %s\n", err)
		}

		client, err := listener.Accept()
		if err != nil {
			log.Fatalln(err)
		}

		handleClient(client, local)
	}

}

func setWinsize(f *os.File, w, h int) {
	syscall.Syscall(syscall.SYS_IOCTL, f.Fd(), uintptr(syscall.TIOCSWINSZ),
		uintptr(unsafe.Pointer(&struct{ h, w, x, y uint16 }{uint16(h), uint16(w), 0, 0})))
}

func server() {
	gssh.Handle(func(s gssh.Session) {
		shell := "sh"
		for _, e := range os.Environ() {
			pair := strings.SplitN(e, "=", 2)
			if pair[0] == "SHELL" {
				shell = pair[1]
			}
		}

		cmd := exec.Command(shell)
		_, winCh, isPty := s.Pty()
		if isPty {
			// cmd.Env = append(cmd.Env, fmt.Sprintf("TERM=%s", ptyReq.Term))
			cmd.Env = os.Environ()
			f, err := pty.Start(cmd)
			if err != nil {
				panic(err)
			}
			go func() {
				for win := range winCh {
					setWinsize(f, win.Width, win.Height)
				}
			}()
			go func() {
				io.Copy(f, s) // stdin
			}()
			io.Copy(s, f) // stdout
			cmd.Wait()
		} else {
			io.WriteString(s, "No PTY requested.\n")
			s.Exit(1)
		}
	})

	publicKeyOption := gssh.PublicKeyAuth(func(ctx gssh.Context, key gssh.PublicKey) bool {
		success := gssh.KeysEqual(key, authorizedKey)
		if !success {
			log.Printf("Public key auth failed")
			log.Printf("Auth key: %v\n", authorizedKey)
			log.Printf("User key: %v\n", key)
		}
		return success
	})

	hostKeyOption := gssh.HostKeyPEM(gprivateKey)

	log.Println("Starting server... Connect via rendezvous sever:")
	log.Printf("  ssh -J %s:%d -p %d debugci@localhost\n", rendezvousEndpoint.Host, rendezvousEndpoint.Port, inspectionEndpoint.Port)
	log.Fatal(gssh.ListenAndServe(":2222", nil, publicKeyOption, hostKeyOption))
}
